# Author:: Gerhard Spitzlsperger
# EMAIL and WWW:: ruby@gs68.de, www.gs68.de
# Copyright:: Copyright (c) 2003 Gerhard Spitzlsperger. All rights reserved.
# License:: LGPL see file LICENSE.
# Version 0.1


require 'GS_Math'
require 'GS_Physics'
require 'GS_Semiconductor'

include Math
include GS_Math
include GS_Physics
include GS_Semiconductor

class FETMOS 
  # the capacities [fF]and external voltages [V] 
  attr_reader   :c
  attr_reader   :phi
  attr_reader   :nodes
  
  # floating gate potential [V] and charge [fC] 
  attr_reader   :phi_fg, :q_fg
  
  # total capacity
  attr_reader   :c_total
  
  # tunneling related oxide thickness [nm] and area [um^2] (total and %drain)
  attr_reader   :t_fto, :a_fto, :d_over

  # control gate related (capacity per area [fF/um^2] and area [um^2]) 
  attr_reader   :c_a_cg, :a_cg
  
  attr_accessor :test

  def initialize
    @nodes = %w(cg sub drain field source) 
  
    # init everything with zero
    @phi_fg = @q_fg = 0.0
    @t_fto = @a_fto = @d_over = 0.0
    @c_a_cg = @a_cg = 0.0
  
    @c = Hash.new
    @phi = Hash.new
  
    @nodes.each { |item| 
      @c[item] = 0.0
      @phi[item] = 0.0      
    }
    
    @fto = FowlerNordheim_Tunneling.new()
  end
  
  # recalculates the total capacity
  def calc_c_total
    @c_total = 0.0
    
    @c.each_value { |item|
      @c_total += item
    }

    return @c_total
  end
  
  # calculate all capacities new
  # for simplicity in the initial version the source and field capacity
  # is set to zero
  def calcCapacities
    d = @d_over / 100.0 # in %
    
    # all capacities are in femto fF
    @c["source"] = 0.0
    @c["field"] = 0.0
    @c["drain"] = EPS_0 * EPS_SiO * @a_fto * d / @t_fto * 1.0e12
    @c["sub"] = EPS_0 * EPS_SiO * @a_fto * (1.0 -d) / @t_fto * 1.0e12
    @c["cg"] = @a_cg * @c_a_cg 	  
  end
  
  # calc the floating gate potential
  def calcPhiFG
    calcCapacities
    calc_c_total
    
    prod = 0.0
    @nodes.each { |item|
      prod += @c[item] * @phi[item] 
    }
    
    prod += @q_fg
    
    @phi_fg = prod / @c_total
  end

  # calc the floating gate potential using an arbitrary charge
  # this function is needed for write/erase simulation
  def calcPhiFG_q(q)
    calcCapacities
    calc_c_total
    
    prod = 0.0
    @nodes.each { |item|
      prod += @c[item] * @phi[item] 
    }
    
    prod += q
    
    prod / @c_total
  end  

  # helper function to verify that a given node exists
  def checkNode( node )
    # prepare for exceptions
    begin   
      if( ! nodes.include?(node) )
           Kernel::raise UnexpectedNodeException, 
	   "Unexpected Node Exception"
      end	 
    end # exceptions  
  end
  
  # get the voltage between the external node and the floating gate
  def getVoltage( node )
    calcPhiFG
   
    checkNode( node ) # raises an exception if the node does't exist
    
    @phi_fg - @phi[node]
  end  
  
  # set the control gate related issues and recalculates FG potential
  # c_a_cg = capacity/area
  # a_cg = area
  def setCG( c_a_cg, a_cg )
    @c_a_cg = c_a_cg
    @a_cg = a_cg
    
    calcPhiFG
  end  
    
  # set the fto related issues and recalculates FG potential
  # t_fto = thickness
  # a_fto = area
  # d_over = drainoverlap  
  def setFTO( t_fto, a_fto, d_over )
    @t_fto = t_fto 
    @a_fto = a_fto 
    @d_over = d_over
    
    calcPhiFG
  end  
  
  # set the floaing gate charge and recalculates FG potential
  def set_q_FG( q_fg )
    @q_fg = q_fg
    
     calcPhiFG
  end  
       
  # set an external nodes potential
  def setPhi( node, value )
    checkNode( node ) # raises an exception if the node does't exist
    
    @phi[node] = value
    
    calcPhiFG
  end  
  
  # calcs the instant drain current through FTO (note the current changes the
  # FG charge and changes therefore rapidly. Result in [A]
  def calcDrainCurrent 
    d = @d_over / 100.0 # in %
    j = @fto.calc( getDrainField ) * @a_fto * d * 1e-8
  end  
    
  def calcSubCurrent 
    d = @d_over / 100.0 # in %
    j = @fto.calc( getSubField ) * @a_fto * (1.0 - d) * 1e-8
  end  

  # calculates the electric field at the drain, assuming capacities
  # and voltages are already calculated. Result in MV/cm
  protected
  def getDrainField
    get_Efield_across_Oxide(@phi_fg - @phi['drain'], @t_fto)
  end  

  protected
  def getSubField
    get_Efield_across_Oxide(@phi_fg - @phi['sub'], @t_fto)
  end     
  
  # simulate writing through substrate
  public
  def simulate
     r = ModEulerDE.new(0, @q_fg, 0.0001, 100, self)
  
     return r
  end   
  
  # the function calc is defined to be able to use the fetmos object
  # in differential equation solvers 
  # !!!!!! should be added dynamically when needed
  # q is the charge in fC, t would be the time but is not needed here 
  public
  def calc ( q, t)
    d = @d_over / 100.0 # in %
    phi_fg =  calcPhiFG_q(q)
    #p "phi_fg= #{phi_fg}"
    #p "q= #{q}"
    
    vdiff = phi_fg - @phi['sub']
    #p "vdiff= #{phi_fg - @phi['sub']}"
    #p "_fto= #{@a_fto}"
    
    # for the moment restrict to substrate tunneling, this is the instant current in [A]
    j = @fto.calc(  get_Efield_across_Oxide(vdiff.abs, @t_fto)) * 
            @a_fto * (1.0 - d) * 1e-8
     
    # decide sign and convert to fA
    if  ( vdiff >= 0.0 )
       j = -1.0 * j*1.0e15
    else
       j = j*1.0e15
    end
     
    p "current #{j} \n"
    p " " 
     
    # return the current
    j
  end
end


# define exceptions
class UnexpectedNodeException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end