# Author:: Gerhard Spitzlsperger
# EMAIL and WWW:: ruby@gs68.de, www.gs68.de
# Copyright:: Copyright (c) 2003 Gerhard Spitzlsperger. All rights reserved.
# License:: LGPL see file LICENSE.
# Version 0.2

# changes 0.2
# * add random number generator for poisson


module GS_Math

class Interpolation
  def initialize(x_data, y_data, type)
    @type = type
    
    # actually only linear is supported
    @type = 'lin'
    
    # it is assumed that the data is ordered
    @x = x_data
    @y = y_data
  end

  def calc(x)
    # if x has the minimum allowed value return it
    begin # exceptions
      if x < @x.first || x > @x.last
        Kernel::raise OutOfRangeException, 
	    "out of interpolation range"
      elsif @x[0] == x
        result = @y[0] 
      else  
        i = 0
        while @x[i] <= x
          i += 1
        end
    
        # now @x[i] is >= x  
        ratio = (x - @x[i-1]) / (@x[i] - @x[i-1])
  
        result = @y[i-1] + ratio * (@y[i] - @y[i-1])
      end # exceptions	
    end
    
    return result
  end  
end


class SimpleEulerDE
  def initialize(x0, y0, xf, n, f)
    @x0 = x0 # starting value for the iteration
    @y0 = y0 # boundry condition
    @xf = xf # final value
    @n = n   # number of points to be calculated
    @f = f   # object through with the first derivative 
             # can be calculated y' = f(y,x) must
	     # implement a method f.calc(y,x)
    
    @h = (@xf - @x0) / (@n - 1)  # step
    
    # array for the result
    @result = Array.new(@n)
    @x = Array.new(@n)
    
    # solve the DE
    init
    
    @i = Interpolation.new(@x, @result, 'lin')
  end

  # interpolation for the result
  def calc(x)
    @i.calc(x)
  end
  
  # do the initialization, will be overridden by the modified and 
  # improved Euler Method for ordinary DE
  protected
  def init
    @result[0] = @y0
    @x[0] = @x0
    
    for i in 1..@n
      @result[i] = @result[i-1] + @h * @f.calc(@result[i-1], @x[i-1])
      @x[i] = @x[i-1] + @h
    end
  end
end 

# the modified Euler method (basically everything beside the solving
# is kept from the ancestor). Now the derivative is taken in the middle of the
# interval not at the left side
class ModEulerDE < SimpleEulerDE
  protected
  def init
    @result[0] = @y0
    @x[0] = @x0
    
    for i in 1..@n
      x_mid = @x[i-1] + @h/2.0
      y_mid = @result[i-1] + @h/2.0 * @f.calc(@result[i-1], @x[i-1])
    
      @result[i] = @result[i-1] + @h * @f.calc(y_mid, x_mid)
      @x[i] = @x[i-1] + @h
    end
  end
end

# the improved Euler method (basically everything beside the solving
# is kept from the ancestor). Now the derivative is taken as the
# average of edges of the interval
class ImprovedEulerDE < SimpleEulerDE
  protected
  def init
    @result[0] = @y0
    @x[0] = @x0
    
    for i in 1..@n
      f = @f.calc(@result[i-1], @x[i-1])
      
      x_h = @x[i-1] + @h
      y_h = @result[i-1] + @h * f
    
      @result[i] = @result[i-1] + @h * (@f.calc(y_h, x_h) + f) / 2.0
      @x[i] = @x[i-1] + @h
    end
  end
end


# uniform distributed random numbers between 0 .. 1 (use 
# the library rand function
class UniformRandomNumbers
  def next_num
    rand
  end
end


class PoissonRandomNumbers
  attr_reader :mean

  def initialize(m)
    reset(m)	
  end
    
  def next_num
    event_counter = -1
    t = 1.0  # time between events
    
    begin
      event_counter += 1
      t *= rand
      #print "#{t}, #{event_counter} \n";
    end while t  > @exp_mean  
    
    #print "done exp_mean was: #{@exp_mean}\n";
    
    return event_counter
  end  
    
  def reset(m)
    @mean = m
    @exp_mean = Math.exp(-1.0 * @mean)
  end
end

class GaussianRandomNumbers
  attr_reader :mean
  attr_reader :sigma

  def initialize(m, s)
    reset(m, s)
  end
    
  def next_num
    if( @one_more_available )
      num = @spare_num
      @one_more_available = false
    else
      num = generate_2_nums
      @one_more_available = true
    end

    return @sigma * num + @mean
  end  
    
  def reset(m, s) 
    @mean = m
    @sigma = s
    
    @one_more_available = false
  end
  
  protected
  def generate_2_nums  # standard gaussian
    r = 0.0
    begin
      h1 = 2.0 * rand - 1.0
      h2 = 2.0 * rand - 1.0
      r = h1 * h1 + h2 * h2
    end while r >= 1.0 || r == 0.0
    
    temp = Math.sqrt(-2.0 * Math.log(r)/r)   
    
    @spare_num = h1 * temp

    return h2 * temp
  end
end



# define exceptions
class OutOfRangeException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end

end