# Author:: Gerhard Spitzlsperger
# EMAIL and WWW:: ruby@gs68.de, www.gs68.de
# Copyright:: Copyright (c) 2003 Gerhard Spitzlsperger. All rights reserved.
# License:: LGPL see file LICENSE.
# Version 0.11


# change 0.11
# * add possibility to plot symbols to GS_FUNCTION
# * add public min/max values to function 

# 0.1 initial version


require 'GS_Style'

class GS_PlotItem
  attr_reader   :type
  attr_accessor :depth
  attr_accessor :baseStyle
  attr_accessor :isVisible

  def initialize
    @depth = 0
    @isVisible = true
  end
end

class GS_BaseAxis < GS_PlotItem
  attr_reader   :major_tick, :minor_tick, :first_tick
  attr_reader   :x1, :y1, :x2, :y2
  attr_reader   :major_tick_pos, :minor_tick_pos
  attr_accessor :tick_size, :minor_tick_size, :x_num_dist, :y_num_dist
  attr_accessor :num_tick_offset
  attr_accessor :ticks_and_numbers

  def initialize
    # Invoke base class initialize function first
    super()
    
    @tick_size = 5
    @minor_tick_size = 3
    @x_num_dist = 15
    @y_num_dist = 5
    @num_tick_offset = 3
    @ticks_and_numbers = true
  end

  def calcLabel(pos, min, max)
    "abstract"
  end  

  def scaleTicks
    "abstract"
  end
  
  def setX(x1, x2, y)
    #d = (x2 - x1) * 0.02
    @x1, @x2 = x1, x2
    
    # if given set new y value
    @y1, @y2 = y, y if(y)
    
    # prepare the tick mark positions
    scaleTicks
  end  
  
  def setY(y1, y2, x)
    #d = (y2 - y1) * 0.02
    @y1, @y2 = y1, y2
    
    # if given set new x value
    @x1, @x2 = x, x if( x )    

    # prepare the tick mark positions
    scaleTicks
  end  
end  

class GS_Axis < GS_BaseAxis
  attr_reader   :omit_0_label
    
  def initialize(type, x1, y1, x2, y2, major_tick, first_tick, omit_0_label)
    super()
    # define type of Axis
    @type = type

    # keep argument settings
    @x1, @y1, @x2, @y2 = x1, y1, x2, y2
    @major_tick, @first_tick, @omit_0_label = major_tick, first_tick, omit_0_label

    # prepare the tick mark positions
    scaleTicks
  end
  
  def scaleTicks  
    # prepare array to hold tickmark positions
    @major_tick_pos = Array.new(0)

    max = (@type =~ /GS_YAxis/) ? @y2 : @x2
    min = (@type =~ /GS_YAxis/) ? @y1 : @x1 
    
    temp = @first_tick

    if(@type =~ /Autoscale/)
       tick_count = 1.2
       @major_tick = 10**(Math.log10((max - min)/ tick_count).floor) 
       
       temp = min + @major_tick 
    end
  
    # prepare for exceptions
    begin   
      i=0
      begin
        @major_tick_pos.push(temp)
        temp += @major_tick
   
        i += 1
        if( i > 50 )
           Kernel::raise AbnormalHighTickCountException, 
	   "too many ticks on axis"
        end	 
      end while temp < max - 0.3*@major_tick  # round off errors
    end # exceptions

  end    
  
  def calcLabel(pos, min, max)
    pos_s = sprintf("%.2f", pos)
  end

end 
  
class GS_Function < GS_PlotItem
  attr_reader   :x1, :x2, :n, :f
  attr_reader   :data
  attr_reader   :min, :max

  def initialize(x1, x2, n, f)  
    # Invoke base class initialize function first
    super()
    
    @type = "GS_Function" 

    @f = f

    @x1, @x2 = x1, x2
    @n = n
    @data = Array.new(0)
    calcValues()
  end
  
  # calcs all the values required
  def calcValues
    step = 1.0 * (x2 - x1) / (n-1)
    temp = x1
    
    i = 1
    while i <= n do
      @data.push(GS_Point.new(temp, function(temp)))   
      
      i += 1
      temp += step
    end  
    
    @max = (@data.map { |p| p.y} ).max
    @min = (@data.map { |p| p.y} ).min
  end
  
  # calculate individual values, this assumes that the object @f implemets
  # a method calc
  def function(x)
    @f.calc(x)
  end   
  
  def pr
    print f.to_s + "\n"
    p "#{@x1} : #{@x2}"
    p "max #{@max}, min #{@min}"
    p "end"
  end   
end

class GS_Line < GS_PlotItem
  attr_reader   :x1, :y1, :x2, :y2
  
  def initialize(x1, y1, x2, y2)  
    # Invoke base class initialize function first
    super()
    
    @type = "GS_Line" 

    @x1 = x1
    @y1 = y1
    @x2 = x2
    @y2 = y2
  end
end

class GS_LogAxis < GS_BaseAxis
    
  def initialize(type, x1, y1, x2, y2)
    super()
    # define type of Axis
    @type = type

    @major_tick = 1.0

    # keep argument settings
    @x1, @x2, @y1, @y2 = x1, x2, y1, y2
    
    scaleTicks
  end
  
  def calcLabel(pos, min, max)
    pos_s = sprintf("1.0e%d", pos)
  end
  
  def scaleTicks
    # prepare array to hold tickmark positions
    @major_tick_pos = Array.new(0)

    max = (@type =~ /GS_LogYAxis/) ? @y2 : @x2
    min = (@type =~ /GS_LogYAxis/) ? @y1 : @x1 
    
    #temp = (@type =~ /Autoscale/) ? min + @major_tick : @first_tick
    temp = min
    
    @minor_tick_pos = [2, 3, 4, 5, 6, 7, 8, 9]
    
    begin
      @major_tick_pos.push(temp)
      temp += @major_tick
    end while temp < max
    
    #p @type, @major_tick_pos
  end    
end

class GS_Point < GS_PlotItem
  attr_reader   :x, :y 
  
  def initialize(x, y)
    # Invoke base class initialize function first
    super()
    
    @type = "GS_Point"

    @x = x
    @y = y
  end
  
  def setX(x)
    @x = x
  end

  def setY(y)
    @y = y
  end  
end

class GS_Rectangle < GS_PlotItem
  attr_reader   :x, :y, :w, :h 
  
  def initialize(x, y, w, h)
    # Invoke base class initialize function first
    super()
    
    @type = "GS_Rectangle"

    @x = x
    @y = y
    @w = w
    @h = h
  end
  
  def setX(x)
    @x = x
  end

  def setY(y)
    @y = y
  end  
end

class GS_Text < GS_PlotItem
  attr_reader   :x, :y 
  attr_accessor :text
  
  def initialize(x, y, text)  
    # Invoke base class initialize function first
    super()
    
    @type = "GS_Text" 

    @x = x
    @y = y
    
    @text = text
  end
end

# special class to hold sums of GS_Functions
class GS_SumFunctionContainer
  attr_reader   :items 

  def initialize 
    @items = Hash.new("not existing")
  end
  
  def calc(x)
    r = 0.0
    @items.each_value { |item| r += item.function(x) }
    
    return r
  end
  
  def append_item(name, item)
     @items.update({name => item})
  end   
end

# special class to hold sums of GS_Functions
class GS_DiffFunctionContainer
  attr_reader   :a, :b   # a and b need to be  GS_Function Objects

  def initialize (a, b)
    @a = a     
    @b = b
    
    if @a.class.to_s != "GS_Function"
      p @a
      Kernel::raise NoGS_FunctionException, "no GS_Function"
    end
    
    if @b.class.to_s != "GS_Function"
      # make a singleton method that returns 0
      def @b.function(x); 0; end;
    end
 
  end
  
  def calc(x)
    r = @a.function(x) - @b.function(x) 
    return r
  end
end



# define exceptions
class AbnormalHighTickCountException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end

class NoGS_FunctionException < RuntimeError
  attr :okToRetry
  def initialize(okToRetry)
    @okToRetry = okToRetry
  end
end